#include <slb_common.h>


/* Global variables */
static unsigned short	debug = 1;

static FILE	*Logf = NULL;
static char	*log_file = "/var/log/slb/slbcd.log";
static int	port = 9100;
static short	ipv4 = 1, ipv6 = 1;
static int	become_daemon = 1;


/*
 * Returns the load
 */
static void lbd_get_load(char *answer, const size_t answer_size)
{
	double avg[3];
	unsigned long load;
	int r;

	snprintf(answer, answer_size, "0");

	r = getloadavg(avg, 3);
	if (r != -1) {
		Log(8, "LOAD: %f %f %f\n", avg[0], avg[1], avg[2]);
		load = avg[0] * 1000;
		snprintf(answer, answer_size, "LOAD load=%lu\n", load);
	}
}

/*
 * Returns info about a port
 * Returns: @answer
 * TODO: Send a notification on server if a port is down
 */
static void lbd_get_port(char *answer, const size_t answer_size, char *s)
{
	char *dump;
	char *addr, *sprotocol;
	int port;
	struct Conn *X;
	int protocol, family;
	struct Conn_split *sp;

	dump = Conn_dump(s, strlen(s));
	Log(3, "%s: cmd=[%s]\n", __FUNCTION__, dump);
	free(dump);

	snprintf(answer, answer_size, "PORT busy=0 msg=\"port is down\"\n");

	sp = Conn_split(s + 5);
	if (!sp) {
		Log(0, "Cannot parse line (%s)!\n", Conn_strerror());
		return;
	}

	addr = Conn_split_get(sp, "addr");
	if (!addr)
		addr = "0.0.0.0";

	port = Conn_split_get_ul(sp, "port", 10);

	protocol = SOCK_STREAM;
	sprotocol = Conn_split_get(sp, "protocol");
	if (sprotocol) {
		if (strcmp(sprotocol, "udp") == 0)
			protocol = SOCK_DGRAM;
	}

	family = Conn_addr_family(addr);
	if (family == -1) {
		Log(0, "ERROR: Cannot find out family from addr [%s] (%s).\n",
			addr, Conn_strerror());
		goto split_free;
	}

	Log(2, "INFO: Trying to bind to %s/%s/%d/%s...\n",
		slb_get_family_type(family), addr, port,
		slb_get_protocol(protocol));
	X = Conn_socket_addr(family, protocol, addr, port);
	if (X == NULL) {
		Log(2, "INFO: Port is taken, so we are OK!\n");
		snprintf(answer, answer_size, "PORT busy=1 msg=\"port seems used!\"\n");
	} else {
		Log(2, "INFO: Port is not taken, not good!\n");
		Conn_close(X);
	}

	split_free:
	Conn_split_free(&sp);
}

static void s_accept(struct Conn *C)
{
	Log(4, "%s: A connection was accepted from [%s/%d]...\n",
		__FUNCTION__, Conn_addr_remote(C), Conn_port_remote(C));
}

static void s_close(struct Conn *C)
{
	Log(5, "%s: A connection will be closed [%s/%d]...\n",
		__FUNCTION__, Conn_addr_remote(C), Conn_port_remote(C));
}

static int s_data_cb(struct Conn *C, char *line)
{
	int len, do_close = 0;
	char ans[128];
	char *dump;
	unsigned int val;

	len = strlen(line);
	if (len == 0)
		return 0;

	if (debug >= 8) {
		dump = Conn_dump(line, len);
		Log(8, "%s: recv: [%s]\n",
			__FUNCTION__, dump);
		free(dump);
	}

	if (strncmp(line, "LOAD", 4) == 0) {
		lbd_get_load(ans, sizeof(ans));
	} else if (strncasecmp(line, "QUIT", 4) == 0) {
		snprintf(ans, sizeof(ans), "Bye!\n");
		do_close = 1;
	} else if (strncmp(line, "PORT", 4) == 0) {
		lbd_get_port(ans, sizeof(ans), line);
	} else if (strncmp(line, "DEBUG ", 6) == 0) {
		val = strtoul(line + 6, NULL, 10);
		Conn_debug(Logf, val);
		snprintf(ans, sizeof(ans), "Main debug level is now %u!\n", val);
	} else if (strncmp(line, "CONNDEBUG ", 10) == 0) {
		val = strtoul(line + 10, NULL, 10);
		Conn_debug(Logf, val);
		snprintf(ans, sizeof(ans), "Conn debug level is now %u!\n", val);
	} else {
		Log(2, "WARN: unknown command!\n");
		snprintf(ans, sizeof(ans),
			"HELP:\n"
			"LOAD - get load\n"
			"QUIT - disconnect\n"
			"PORT - get info about a port\n"
			"\n");
	}

	dump = Conn_dump(ans, strlen(ans));
	Log(8, "Send back: [%s]\n", dump);
	free(dump);
	Conn_enqueue(C, ans, strlen(ans));

	if (do_close == 1)
		Conn_close(C);

	return 0;
}

static void s_data(struct Conn *C)
{
	Conn_for_every_line(C, s_data_cb);
}

static void s_error(struct Conn *C)
{
	Log(1, "%s: cid=%llu (%s)\n",
		__FUNCTION__, Conn_getid(C), Conn_errno(C));
}

int main(void)
{
	struct Conn *I4 = NULL, *I6 = NULL;
	int ret;
	struct timeval start;
	FILE *f;
	char *r;
	char buf[128];

	f = fopen("/etc/slb/slbcd.conf", "r");
	if (!f) {
		Log(0, "\tCannot open /etc/slb/slbcd.conf file (%s)! Ignore it.\n",
			strerror(errno));
	} else {
		while (fgets(buf, sizeof(buf) - 1, f)) {
			r = buf + strlen(buf) - 1;
			while ((r >= buf) && ((*r == ' ') || (*r == '\t') || (*r == '\n') || (*r == '\r'))) {
				*r = '\0';
				r--;
			}

			if ((strlen(buf) == 0) || (buf[0] == '#') || (buf[0] == ';'))
				continue;

			if (strncmp(buf, "debug", 5) == 0) {
				r = buf + 5;
				while ((*r == ' ') || (*r == '\t') || (*r == '='))
					r++;
				debug = strtoul(r, NULL, 10);
				continue;
			}

			if (strncmp(buf, "daemon", 5) == 0) {
				r = buf + 5;
				while ((*r == ' ') || (*r == '\t') || (*r == '='))
					r++;
				become_daemon = strtol(r, NULL, 10);
				continue;
			}

			Log(0, "Error parsing line [%s]. Ignore it.\n", buf);
		}
		fclose(f);
	}

	Logf = fopen(log_file, "a+");
	if (!Logf) {
		fprintf(stderr, "Cannot open log file [%s] [%s]\n",
			log_file, strerror(errno));
		return 1;
	}
	if (debug > 0)
		setlinebuf(Logf);

	if (become_daemon == 1)
		daemon(0, 0);

	/* set library output debug */
	Conn_debug(Logf, debug);

	gettimeofday(&start, NULL);
	Log(0, "Starting...\n");
	Log(0, "\tPort=%d\n", port);
	Log(0, "\tLogFile=%s Debug=%d\n", log_file, debug);
	Log(0, "\tipv4=%s ipv6=%s\n", (ipv4 == 1) ? "on" : "off", (ipv6 == 1) ? "on" : "off");

	ret = Conn_init(0);
	if (ret == -1) {
		Log(0, "%s", Conn_strerror());
		return 1;
	}

	if (ipv4 == 1) {
		Log(9, "Try to register IPv4 socket...\n");
		I4 = Conn_socket(PF_INET, SOCK_STREAM, port);
		if (!I4)
			Log(1, "Cannot bind on ipv4 socket [%s].\n", Conn_strerror());
	}

	if (ipv6 == 1) {
		Log(9, "Try to register IPv6 socket...\n");
		I6 = Conn_socket(PF_INET6, SOCK_STREAM, port);
		if (!I6)
			Log(1, "Cannot bind on ipv6 socket [%s].\n", Conn_strerror());
	}

	if ((I6 == NULL) && (I4 == NULL)) {
		Log(0, "Cannot bind!\n");
		return 1;
	}

	Conn_accept_cb = s_accept;
	Conn_data_cb = s_data;
	Conn_close_cb = s_close;
	Conn_error_cb = s_error;

	while (1) {
		ret = Conn_poll(-1);
		if (ret == -1) {
			Log(0, "Error calling Conn_poll [%s]!\n", Conn_strerror());
			return 1;
		}
	}

	Log(0, "Finish!\n\n");

	return 0;
}
